local S = core.get_translator(core.get_current_modname())

local function create_quaternion(axis, angle)
	local half_angle = angle / 2
	local sin_half_angle = math.sin(half_angle)
	local cos_half_angle = math.cos(half_angle)

	return {
		w = cos_half_angle,
		x = axis.x * sin_half_angle,
		y = axis.y * sin_half_angle,
		z = axis.z * sin_half_angle,
	}
end

local function rotate_by_quaternion(v, q)
	local xx = q.x * q.x
	local yy = q.y * q.y
	local zz = q.z * q.z
	local xy = q.x * q.y
	local xz = q.x * q.z
	local yz = q.y * q.z
	local wx = q.w * q.x
	local wy = q.w * q.y
	local wz = q.w * q.z

	return {
		x = (1 - 2 * (yy + zz)) * v.x + (2 * (xy - wz)) * v.y + (2 * (xz + wy)) * v.z,
		y = (2 * (xy + wz)) * v.x + (1 - 2 * (xx + zz)) * v.y + (2 * (yz - wx)) * v.z,
		z = (2 * (xz - wy)) * v.x + (2 * (yz + wx)) * v.y + (1 - 2 * (xx + yy)) * v.z,
	}
end

local definitions = {
	stone = {
		damage = {fleshy=4,knockback=8},
		craft_amount = 10,
	},
	steel = {
		damage = {fleshy=6,knockback=3},
		craft_amount = 30,
	},
	bronze = {
		damage = {fleshy=6,knockback=3},
		craft_amount = 30,
	},
	gold = {
		damage = {fleshy=5,knockback=10},
		craft_amount = 30,
	},
	mese = {
		damage = {fleshy=8,knockback=2},
		craft_amount = 60,
	},
	diamond = {
		damage = {fleshy=7,knockback=2},
		craft_amount = 50,
	},
}

local shot_bullet = {
	timer = 0,
	initial_properties = {
		static_save = false,
		pointable = false,
		physical = false,
		hp_max = 420,
		visual = "wielditem",
		visual_size = {x = 0.75, y = 0.75},
		lastpos = {},
		selectionbox = {-0.5, -0.5, -0.0025, 0.5, 0.5, 0.0025, rotate=true},
	},
	_rotation_timer = 0,
	_ricochet = 0,
	on_activate = function (self, staticdata, dtime_s)
		self._old_pos = self.object:get_pos()
	end,
	on_punch = function (self, puncher, time_from_last_punch, tool_capabilities, dir, damage)
		if self._attached and puncher:is_player() then
			local inv = puncher:get_inventory()
			if inv:room_for_item("main",self._drop) then
				inv:add_item("main",self._drop)
				self.object:remove()
				return
			end
		end
	end,
	_attach = function(self,node_pos,hit_pos,pointed_thing)
		self._ricochet = self._ricochet + 1
		self._attached = true
		self._attach_pos = node_pos
		
		-- Move the bullet slightly above the hit position
		local new_pos = vector.add(hit_pos, vector.multiply(((pointed_thing.intersection_normal or vector.new())), 0.3)) -- adjust the 0.1 value as needed
		self.object:set_pos(new_pos)
		
	
		
		local o = self.object
		o:set_pos(new_pos)
		o:set_properties({pointable = true})
		o:set_acceleration(vector.new())
		o:set_velocity(vector.new())
	end,
	_unattach = function(self)
		self._attached = false
		local o = self.object
		o:set_properties({pointable = false})
		o:set_acceleration(vector.new(0,-9.81,0))
	end,
on_step = function(self, dtime, moveresult)
	if self.owner == nil then
		self.object:remove()
		return
	end
	
	self.timer = self.timer + dtime
	if self.timer > 900 then
		self.object:remove()
		return
	end
	
	if self._attached then 
		if core.get_node(self._attach_pos).name == "air" then
			self:_unattach()
		end
		return
	end
	
	self._rotation_timer = self._rotation_timer + dtime
	if self._rotation_timer >= 0.2 then
		self._rotation_timer = 0
		local rotation_speed = math.pi * 3  -- Rotation speed (radians per second)
		local current_rotation = self.object:get_rotation()

		-- Update rotation
		current_rotation.z = current_rotation.z + rotation_speed * -0.2
		self.object:set_rotation(current_rotation)
	end

	local pos = self.object:get_pos()

	for pointed_thing in core.raycast(self._old_pos, self.object:get_pos(), true, true) do
		local hit_pos = pointed_thing.intersection_point

		if pointed_thing.type == "node" then
			local node_pos = pointed_thing.under
			local node = core.get_node(node_pos)
			local node_def = core.registered_nodes[node.name]
			if node_def and node_def.walkable then
				core.check_for_falling(node_pos)

				local nodedef = minetest.registered_nodes[node.name]
				if nodedef and nodedef.sounds and (nodedef.sounds.dig or nodedef.sounds.dug) then
					core.sound_play((nodedef.sounds.dig or nodedef.sounds.dug).name, {pos = hit_pos}, true)
				end

				self:_attach(node_pos,hit_pos,pointed_thing)
				return
			end
		elseif pointed_thing.type == "object" and (
				(pointed_thing.ref:is_player() and (pointed_thing.ref:get_player_name() ~= self.owner or self._ricochet > 0))
				or
				(pointed_thing.ref:get_luaentity() and pointed_thing.ref:get_properties().physical and pointed_thing.ref:get_luaentity().name ~= '__builtin:item')
			) then
			local actualDamage = self.damage or {fleshy = 1}
			local damage = {}
			local crit = self.crit or 0
			local critEffc = self.critEffc or 1
			local owner = core.get_player_by_name(self.owner)
			local hit_texture = "hit.png"
			local dps = self.dps or 0
			local skill = self.skill_value or 1

			for k, dmg in pairs(actualDamage) do
				damage[k] = actualDamage[k]
			end

			if owner then accuracy.hit(owner) end
			-- put bullet in front of object before punching in order to get correct engine knockback
			self.object:set_pos(self._old_pos)
			pointed_thing.ref:punch(
				self.object,
				1.0,
				{
					full_punch_interval = 1.0,
					damage_groups = damage
				},
				nil
			)

			self.object:remove()
			return
		end
	end

	local velocity = self.object:get_velocity()
	local air_density = 1.225  -- kg/m^3 (standard atmospheric density)
	local drag_coefficient = 0.2  -- Example for a sphere, adjust for your entity
	local area = 0.1  -- Cross-sectional area, adjust for your entity
	local drag_force_magnitude = 0.5 * air_density * vector.length(velocity)^2 * drag_coefficient * area
	-- Calculate the drag force vector
	local drag_force = vector.multiply(vector.normalize(velocity), -drag_force_magnitude)
	-- Apply the drag force to the velocity
	local new_velocity = vector.add(velocity, vector.multiply(drag_force, dtime))
	self.object:set_velocity(new_velocity)
	
	self._old_pos = self.object:get_pos()
end
}
core.register_entity("shuriken:shot_bullet", shot_bullet)

local function launch_projectile(itemstack, user)


end
local player_cooldowns = {}
core.register_on_joinplayer(function(player, last_login) player_cooldowns[player:get_player_name()] = 0 end)
core.register_on_leaveplayer(function(player, timed_out) player_cooldowns[player:get_player_name()] = nil end)
local yeet = function(itemstack, user, pointed_thing)
	if user and user:is_player() then
		local player_name = user:get_player_name()
		local current_time = os.clock()
		if current_time < player_cooldowns[player_name] then
			return
		end
		player_cooldowns[player_name] = current_time + 0.3
	
		local pos = user:get_pos()
		local dir = user:get_look_dir()
		local svertical = user:get_look_vertical()
		
		core.sound_play("shuriken_throw", {pos = pos}, true)
		pos.y = pos.y + 1.45

		local obj = core.add_entity(pos, "shuriken:shot_bullet")
		obj:set_properties({textures = {itemstack:get_name()}})
		local ent = obj:get_luaentity()

		local initial_speed = 20
		ent.owner = user:get_player_name()
		ent.damage = definitions[itemstack:get_definition()._variant].damage
		ent._drop = itemstack:get_name()

		local acc = 0.15 + accuracy.get(user)*math.pi*0.5
		accuracy.reset(user)
		-- maximum deviation in radians
		local max_deviation = acc

		-- Apply random deviation within a circular cone
		local angle = math.random() * max_deviation
		local axis = vector.normalize(vector.new(math.random() - 0.5, math.random() - 0.5, math.random() - 0.5))
		local quaternion = create_quaternion(axis, angle)
		local deviated_dir = rotate_by_quaternion(dir, quaternion)

		-- Calculate initial velocity
		local velocity = vector.multiply(deviated_dir, initial_speed)
		
		obj:set_velocity(velocity)
		obj:set_acceleration({x = 0, y = -9.81, z = 0})
		
		if user:is_player() then
			local yaw = user:get_look_horizontal()
			obj:set_yaw(math.atan2(deviated_dir.z, deviated_dir.x))
		end
		
		itemstack:take_item()
	end
	return itemstack
end

local shuriken_recipe = function(variant, ingredient)
	accuracy.register("shuriken:"..variant)
	core.register_craft({
		output = "shuriken:"..variant.." "..definitions[variant].craft_amount,
		recipe = {
			{"", ingredient, ""},
			{ingredient, "", ingredient},
			{"", ingredient, ""},
		}
	})
	core.register_craftitem("shuriken:"..variant, {
		description = S(variant:gsub("^%l", string.upper).." shuriken"),
		range = 0,
		stack_max= 200,
		wield_scale = {x=0.6,y=0.6,z=0.5},
		inventory_image = "shuriken_"..variant..".png",
		_variant = variant,
		on_use = yeet,
	})
end

if minetest.get_modpath("mcl_core") then
	shuriken_recipe("stone","mcl_core:cobble")
	shuriken_recipe("steel","mcl_core:iron_ingot")
	shuriken_recipe("bronze","mcl_core:bronze_ingot")
	shuriken_recipe("gold","mcl_core:gold_ingot")
	shuriken_recipe("diamond","mcl_core:diamond")
else
	shuriken_recipe("stone","default:cobble")
	shuriken_recipe("steel","default:steel_ingot")
	shuriken_recipe("bronze","default:bronze_ingot")
	shuriken_recipe("gold","default:gold_ingot")
	shuriken_recipe("mese","default:mese_crystal")
	shuriken_recipe("diamond","default:diamond")
end